#' Calculate correlation value from discretized distribution
#'
#' @description  For use in simulations. Use to calculate rho or tau from
#'   discretized data generated from the standard bivariate normal distribution.
#'
#' @details Computes the distribution function of the bivariate normal
#'   distribution with correlation \code{r} and thresholds \code{cutx} and
#'   \code{cuty}. Estimates the correlation (rho or tau) from the simulated
#'   distribution.
#'
#' @param cutx,cuty Vector of thresholds for x,y values.
#' @param r Value to discretize.
#'
#' @return Discrete value of rho or tau.
#'
#' @name discrete_r
#'
#' @seealso
#' \code{\link{cluscor}},
#' \code{\link{cluscor.test}}
#'
#' @examples
#' threshx <- c(-Inf,-3,-2,-1,-0.5,0.5,1,2,3,Inf)
#' threshy <- threshx
#' discrete_rho(threshx,threshy,r=0.4)
#' discrete_tau(threshx,threshy,r=0.4)
NULL

#' @rdname discrete_r
#' @export
discrete_rho <- function(cutx, cuty, r) {
  p <- matrix(NA, length(cutx) - 1, length(cuty) - 1)
  for (i in 1:(length(cutx) - 1)) {
    for (j in 1:(length(cuty) - 1)) {
      p[i, j] <- mvtnorm::pmvnorm(lower = c(cutx[i], cuty[j]),
                                  upper = c(cutx[i + 1], cuty[j + 1]),
                                  mean = c(0, 0),
                                  corr = matrix(c(1, r, r, 1), 2, 2))
    }
  }
  px <- apply(p, 1, sum)
  py <- apply(p, 2, sum)
  x <- 1:nrow(p)
  y <- 1:ncol(p)
  fdot.x <- fdot(x, px)
  fdot.y <- fdot(y, py)
  vx.fdot <- var.dot(fdot.x, px)
  vy.fdot <- var.dot(fdot.y, py)
  cov.fdot <- cov.dot(fdot.x, fdot.y, p)
  rho <- cov.fdot/sqrt(vx.fdot * vy.fdot)
  return(rho)
}

#' @rdname discrete_r
#' @export
discrete_tau <- function(cutx, cuty, r) {
  p <- matrix(NA, length(cutx) - 1, length(cuty) - 1)
  for (i in 1:(length(cutx) - 1)) {
    for (j in 1:(length(cuty) - 1)) {
      p[i, j] <- mvtnorm::pmvnorm(lower = c(cutx[i], cuty[j]),
                                  upper = c(cutx[i + 1], cuty[j + 1]),
                                  mean = c(0, 0),
                                  corr = matrix(c(1, r, r, 1), 2, 2))
    }
  }
  px <- apply(p, 1, sum)
  py <- apply(p, 2, sum)
  x <- 1:nrow(p)
  y <- 1:ncol(p)
  fx <- f(x, px)
  fy <- f(y, py)
  fxy <- f2(x, y, p, close = "xy")
  fxminusy <- f2(x, y, p, close = "y")
  fyminusx <- f2(x, y, p, close = "x")
  fminusxy <- f2(x, y, p, close = "none")

  fdot.x <- fdot(x, px)
  fdot.y <- fdot(y, py)
  fdot.xy <- fdot2(x, y, p, "both")
  fdot.xyminus <- fdot2(x, y, p, "x")
  fdot.xminusy <- fdot2(x, y, p, "y")
  num <- 1 - 2 * matrix(fdot.x, length(x), length(y)) -
    2 * matrix(fdot.y, length(x), length(y), byrow = T) + 4 * fdot.xy
  num <- sum(num * p)
  den <- 1 + sum(2 * (fdot.x - fx) * px) + sum(2 * (fdot.y - fy) * py) +
    sum((fxy - fxminusy - fyminusx + fminusxy) * p)
  gamma <- num/den
  tau <- num
  denx <- 1 + sum(2 * (fdot.x - fx) * px)
  deny <- 1 + sum(2 * (fdot.y - fy) * py)
  taub <- tau/sqrt(denx * deny)
  return(c(tau = tau, gamma = gamma, taub = taub))
}

