wcr_rho <- function(x, y, id, order = 1, fixed = TRUE, dist = "normal",
                    conf.level = 0.95, r = 0, ...) {

  alpha <- 1 - conf.level

  ## Subset to complete cases
  if (any(is.na(x) | is.na(y))){
    ind <- which(is.na(x) | is.na(y))
    x <- x[-ind]
    y <- y[-ind]
    id <- id[-ind]
  }
  ## Convert to numeric
  if(is.factor(x)){x <- as.numeric(x)}
  if(is.factor(y)){y <- as.numeric(y)}
  if(is.character(x)){x <- as.numeric(as.factor(x))}
  if(is.character(y)){y <- as.numeric(as.factor(y))}
  ## Convert to character
  if(!is.character(id)){id <- as.character(id)}

  # all options
  METHOD <- "Clustered Spearman's Rho test"

  pval <- se <- numeric()

  n <- length(x)
  n.clust <- length(unique(id))
  # find spearman rho estimate
  x2 <- x[order(id)]
  y2 <- y[order(id)]
  id2 <- sort(id)

  # spearman's rho and variances based cluster-specific cdf
  dumx <- outer(x2, x2, "<=")

  dumx.dot <- 0.5 * outer(x2, x2, "<=") + 0.5 * outer(x2, x2, "<")
  x.mean.dot <- aggrMean(dumx.dot, id2)
  cdfx.dot <- colMeans(x.mean.dot)

  dumy <- outer(y2, y2, "<=")
  y.mean <- aggrMean(dumy, id2)
  cdfy <- colMeans(y.mean)

  dumy.dot <- 0.5 * outer(y2, y2, "<=") + 0.5 * outer(y2, y2, "<")
  y.mean.dot <- aggrMean(dumy.dot, id2)
  cdfy.dot <- colMeans(y.mean.dot)

  cov.wcr <- mean(tapply(cdfx.dot * cdfy.dot, id2, mean)) - 1/4
  vx.wcr <- mean(tapply(cdfx.dot^2, id2, mean)) - 1/4
  vy.wcr <- mean(tapply(cdfy.dot^2, id2, mean)) - 1/4
  rho.wcr <- cov.wcr/sqrt(vx.wcr * vy.wcr)

  # first projection of the U-stat
  x22 <- rep(x2, n)
  y22 <- rep(y2, each = n)
  xy.mean.dot <- aggrMean((0.5 * outer(x2, x22, "<=") + 0.5 * outer(x2, x22, "<")) * (0.5 * outer(y2, y22, "<=") + 0.5 * outer(y2, y22, "<")), id2)
  cdfxy.dot <- colMeans(xy.mean.dot)
  cdfxy.dot <- matrix(cdfxy.dot, nr = n, nc = n)

  psi.1a <- (2 * cdfx.dot * cdfy.dot + 2 * rowMeans(cdfxy.dot) + 2 * colMeans(cdfxy.dot) + 2 - 2 * cdfx.dot - 2 * cdfy.dot)/6 - 1/4 - cov.wcr
  psi.10 <- (2 * cdfx.dot * cdfy.dot - cdfx.dot - cdfy.dot + 2)/6 - 1/4

  psi.1a.bar <- tapply(psi.1a, id2, mean)
  psi.10.bar <- tapply(psi.10, id2, mean)

  psi.x.1a <- (cdfx.dot^2 + 2 * rowMeans(sweep(dumx.dot, 2, cdfx.dot, "*")))/3 - 1/4 - vx.wcr
  psi.x.1a.bar <- tapply(psi.x.1a, id2, mean)

  psi.y.1a <- (cdfy.dot^2 + 2 * rowMeans(sweep(dumy.dot, 2, cdfy.dot, "*")))/3 - 1/4 - vy.wcr
  psi.y.1a.bar <- tapply(psi.y.1a, id2, mean)

  habar.mat.1 <- cbind(psi.1a.bar, psi.x.1a.bar, psi.y.1a.bar)
  var.ha.mat.1 <- 9 * t(habar.mat.1) %*% habar.mat.1/(n.clust^2)

  h0bar.mat.1 <- cbind(psi.10.bar, psi.x.1a.bar, psi.y.1a.bar)
  var.h0.mat.1 <- 9 * t(h0bar.mat.1) %*% h0bar.mat.1/(n.clust^2)


  if (!fixed) {
    delta <- rho.wcr * c(1/cov.wcr, -1/(2 * vx.wcr), -1/(2 * vy.wcr))
    delta <- matrix(delta, 1, 3)
  }
  if (fixed) {
    ratio <- rho.wcr/cov.wcr
  }


  if (order == 1) {
    METHOD <- paste(METHOD, "using a First Order Approximation of Variance",
                    sep = " ")
    if (!fixed) {
      # estimated denominator
      est.var.rho.0.1 <- delta %*% var.h0.mat.1 %*% t(delta)
      est.var.rho.a.1 <- delta %*% var.ha.mat.1 %*% t(delta)

      v1 <- rho.wcr^2/est.var.rho.0.1
      se <- sqrt(c(est.var.rho.0.1, est.var.rho.a.1))

      if (dist == "normal") {
        pval <- 1 - pchisq(v1, 1)
        ci <- ci(rho.wcr, se[2], alpha = alpha, dist = "normal")
      }
      if (dist == "t") {
        pval <- 2 * (1 - pt(abs(rho.wcr/sqrt(est.var.rho.0.1)), n.clust - 2))
        ci <- ci(rho.wcr, se[2], alpha = alpha, dist = "t", df = n.clust - 2)
      }

    }  # estimated denominator
    if (fixed) {
      # fixed denominator
      est.var.rho.fix.0.1 <- var.h0.mat.1[1, 1] * (ratio^2)
      est.var.rho.fix.a.1 <- var.ha.mat.1[1, 1] * (ratio^2)

      v1 <- rho.wcr^2/est.var.rho.fix.0.1
      se <- sqrt(c(est.var.rho.fix.0.1, est.var.rho.fix.a.1))

      if (dist == "normal") {
        pval <- 1 - pchisq(v1, 1)
        ci <- ci(rho.wcr, se[2], alpha = alpha, dist = "normal")
      }
      if (dist == "t") {
        pval <- 2 * (1 - pt(abs(rho.wcr/sqrt(est.var.rho.fix.0.1)), n.clust - 2))
        ci <- ci(rho.wcr, se[2], alpha = alpha, dist = "t", df = n.clust - 2)
      }

    }  # fixed denominator
  }
  if (order == 2) {
    METHOD <- paste(METHOD, "using a Second Order Approximation of Variance",
                    sep = " ")
    cdfx.k.dot <- aggrMean(sweep(dumx.dot, 2, cdfx.dot, "*"), id2)
    cdfx.k.dot <- aggrMean(t(cdfx.k.dot), id2)
    cdfx.dot.mat <- (matrix(cdfx.dot, n, n))
    F.minx <- dumx * cdfx.dot.mat + (1 - dumx) * t(cdfx.dot.mat)
    F.minx <- aggrMean(t(aggrMean(F.minx, id2)), id2)

    psi.x.2.a.mat <- 1/3 * ((cdfx.k.dot) + (t(cdfx.k.dot)) + (1 - (F.minx))) - 1/4

    cdfy.k.dot <- aggrMean(sweep(dumy.dot, 2, cdfy.dot, "*"), id2)
    cdfy.k.dot <- aggrMean(t(cdfy.k.dot), id2)
    cdfy.dot.mat <- (matrix(cdfy.dot, n, n))

    F.miny <- dumy * cdfy.dot.mat + (1 - dumy) * t(cdfy.dot.mat)
    F.miny <- aggrMean(t(aggrMean(F.miny, id2)), id2)

    psi.y.2.a.mat <- 1/3 * ((cdfy.k.dot) + (t(cdfy.k.dot)) + (1 - (F.miny))) - 1/4

    cdfy.kx.dot <- aggrMean(t(aggrMean(sweep(dumx.dot, 2, cdfy.dot, "*"), id2)), id2)
    cdfx.ky.dot <- aggrMean(t(aggrMean(sweep(dumy.dot, 2, cdfx.dot, "*"), id2)), id2)

    cdfx.cdfy.dot <- outer(tapply(cdfx.dot, id2, mean), tapply(cdfy.dot, id2, mean), "+")

    cdfxy.dot.means.a <- aggrMean(t(aggrMean(cdfxy.dot, id2)), id2)
    cdfxy.dot.means.o <- outer(tapply(cdfx.dot, id2, mean), tapply(cdfy.dot, id2, mean), "*")

    psi.xy.2.a.mat <- 1/6 * ((cdfx.ky.dot) + (t(cdfx.ky.dot)) + (cdfy.kx.dot) + (t(cdfy.kx.dot)) + 2 - (cdfx.cdfy.dot) - (t(cdfx.cdfy.dot)) + (cdfxy.dot.means.a) +
                               (t(cdfxy.dot.means.a))) - 1/4
    psi.xy.2.0.mat <- 1/6 * ((cdfx.ky.dot) + (t(cdfx.ky.dot)) + (cdfy.kx.dot) + (t(cdfy.kx.dot)) + 2 - (cdfx.cdfy.dot) - (t(cdfx.cdfy.dot)) + (cdfxy.dot.means.o) +
                               (t(cdfxy.dot.means.o))) - 1/4

    h2.xy.a.mat <- psi.xy.2.a.mat - matrix(psi.1a.bar, n.clust, n.clust, byrow = T) - matrix(psi.1a.bar, n.clust, n.clust, byrow = F) - cov.wcr
    h2.xy.a.vec <- gdata::upperTriangle(h2.xy.a.mat)

    h2.xy.0.mat <- psi.xy.2.0.mat - matrix(psi.10.bar, n.clust, n.clust, byrow = T) - matrix(psi.10.bar, n.clust, n.clust, byrow = F)
    h2.xy.0.vec <- gdata::upperTriangle(h2.xy.0.mat)

    h2.x.a.mat <- psi.x.2.a.mat - matrix(psi.x.1a.bar, n.clust, n.clust, byrow = T) - matrix(psi.x.1a.bar, n.clust, n.clust, byrow = F) - vx.wcr
    h2.x.a.vec <- gdata::upperTriangle(h2.x.a.mat)

    h2.y.a.mat <- psi.y.2.a.mat - matrix(psi.y.1a.bar, n.clust, n.clust, byrow = T) - matrix(psi.y.1a.bar, n.clust, n.clust, byrow = F) - vy.wcr
    h2.y.a.vec <- gdata::upperTriangle(h2.y.a.mat)

    h2bar.a.mat <- cbind(h2.xy.a.vec, h2.x.a.vec, h2.y.a.vec)
    h2bar.0.mat <- cbind(h2.xy.0.vec, h2.x.a.vec, h2.y.a.vec)

    var.h2.a.mat <- t(h2bar.a.mat) %*% h2bar.a.mat/dim(h2bar.a.mat)[1]
    var.h2.0.mat <- t(h2bar.0.mat) %*% h2bar.a.mat/dim(h2bar.0.mat)[1]

    var.a.order2 <- var.ha.mat.1 + 18 * var.h2.a.mat/(n.clust * (n.clust - 1))
    var.0.order2 <- var.h0.mat.1 + 18 * var.h2.0.mat/(n.clust * (n.clust - 1))

    if (!fixed) {
      # estimated denominator
      est.var.rho.0.2 <- delta %*% var.a.order2 %*% t(delta)
      est.var.rho.a.2 <- delta %*% var.0.order2 %*% t(delta)

      v3 <- rho.wcr^2/est.var.rho.0.2
      se <- sqrt(c(est.var.rho.0.2, est.var.rho.a.2))

      if (dist == "normal") {
        pval <- 1 - pchisq(v3, 1)
        ci <- ci(rho.wcr, se[2], alpha = alpha, dist = "normal")
      }
      if (dist == "t") {
        pval <- 2 * (1 - pt(abs(rho.wcr/sqrt(est.var.rho.0.2)), n.clust - 2))
        ci <- ci(rho.wcr, se[2], alpha = alpha, dist = "t", df = n.clust - 2)
      }

    }  # estimated denominator
    if (fixed) {
      # fixed denominator
      est.var.rho.fix.0.2 <- var.0.order2[1, 1] * (ratio^2)
      est.var.rho.fix.a.2 <- var.a.order2[1, 1] * (ratio^2)

      v1.h2 <- rho.wcr^2/est.var.rho.fix.0.2
      se <- sqrt(c(est.var.rho.fix.0.2, est.var.rho.fix.a.2))

      if (dist == "normal") {
        pval <- 1 - pchisq(v1.h2, 1)
        ci <- ci(rho.wcr, se[2], alpha = alpha, dist = "normal")
      }
      if (dist == "t") {
        pval <- 2 * (1 - pt(abs(rho.wcr/sqrt(est.var.rho.fix.0.2)), n.clust - 2))
        ci <- ci(rho.wcr, se[2], alpha = alpha, dist = "t", df = n.clust - 2)
      }

    }  # fixed denominator
  }
  if(!is.na(ci$ciL) & !is.na(ci$ciU)) {
    if(ci$ciU > 1) {
      warning(paste0(
        "The upper limit of the ", (1 - alpha) * 100,"% confidence interval is
      truncated at 1."
      ))
      ci$ciU <- 1
    }
    if(ci$ciL < -1) {
      warning(paste0(
        "The lower limit of the ", (1 - alpha) * 100,"% confidence interval is
      truncated at -1."
      ))
      ci$ciL <- -1
    }
    if(pval < alpha & ci$ciL < 0 & ci$ciU > 0) {
      warning(paste0(
        "The ", (1 - alpha) * 100,"% confidence interval is not consistent with
        statistical testing.\n"
      ))
    }
  }
  names(rho.wcr) <- names(r) <- "rho"
  results <- list(statistic = rho.wcr, p.value = pval, conf.int = ci,
                  n.clust = n.clust, nobs = n, method = METHOD,
                  dist = dist, fixed = fixed, null.value = r,
                  x = x, y = y, id = id)
  class(results) <- "cluscor"
  return(results)
}
